import logging
import urllib.parse
from collections.abc import Iterator
from typing import Any

import ratelimit
import requests
import wikidata.entity

REQUEST_SESSION = None  # TODO?


def concept_uri(obj: wikidata.entity.Entity) -> urllib.parse.ParseResult:
    if obj.id.startswith('P'):
        return urllib.parse.urlparse(f'http://www.wikidata.org/prop/direct/{obj.id}')
    if obj.id.startswith('Q'):
        return urllib.parse.urlparse(f'http://www.wikidata.org/entity/{obj.id}')

    msg = f'Object id scheme not supported: {obj.id}'
    raise ValueError(msg)


def format_value_for_triple_request(obj: Any, prefer_obj=False) -> str:
    if obj is None:
        return ''
    if isinstance(obj, str):
        return f'"{obj}"'
    if isinstance(obj, urllib.parse.ParseResult):
        return (
            obj.geturl()
            if prefer_obj
            else format_value_for_triple_request(obj.geturl())
        )
    if isinstance(obj, wikidata.entity.Entity):
        uri = concept_uri(obj)
        return format_value_for_triple_request(uri, True)

    msg = f'Type cannot be formatted: {type(obj)}'
    raise TypeError(msg)


@ratelimit.sleep_and_retry
def fetch_by_url(url: str, headers: dict[str, str]):
    logging.debug('Fetching: %s', url)
    if REQUEST_SESSION is None:
        msg = 'REQUEST_SESSION must be set, before calling fetch_by_url'
        raise RuntimeError(msg)
    response = REQUEST_SESSION.get(url, headers=headers)
    if not response.status_code.ok:
        logging.error('Got %s error message: %s', response.status_code, response.text)
        return None
    return response


ITEMS_PER_PAGE = 'http://www.w3.org/ns/hydra/core#itemsPerPage'
TOTAL_ITEMS = 'http://www.w3.org/ns/hydra/core#totalItems'


def fmt_params(subject: Any, predicate: Any, object_: Any) -> dict[str, str | int]:
    entities = [x for x in [subject, predicate, object_] if x]
    if len(entities) == 0:
        msg = 'There are no entities for this query!'
        raise RuntimeError(msg)
    return {
        'subject': format_value_for_triple_request(subject, prefer_obj=True),
        'predicate': format_value_for_triple_request(predicate, prefer_obj=True),
        'object': format_value_for_triple_request(object_, prefer_obj=True),
        'page': 1,
    }


def get_triples_count(
    subject: Any = None,
    predicate: Any = None,
    object_: Any = None,
) -> dict[str, Any]:
    """Fetches first page in order to determine amount of items."""
    params = fmt_params(subject, predicate, object_)
    url: str = (
        requests.Request(url='https://query.wikidata.org/bigdata/ldf', params=params)
        .prepare()
        .url
    )
    response = fetch_by_url(url, headers={'accept': 'application/ld+json'})
    if response is None:
        return {
            'items_per_page': 0,
            'items_total': 0,
            'num_pages': 0,
        }
    json_data = response.json()
    for item in json_data['@graph']:
        if TOTAL_ITEMS in item:
            return {
                'items_per_page': item[ITEMS_PER_PAGE],
                'items_total': item[TOTAL_ITEMS],
                'num_pages': int((item[TOTAL_ITEMS] - 1) / item[ITEMS_PER_PAGE] + 1),
            }

    msg = 'Could not determine triple count'
    raise RuntimeError(msg)


def get_triples_internal(subject: Any, predicate: Any, object_: Any) -> Iterator[dict]:
    params = fmt_params(subject, predicate, object_)
    pagination_data = get_triples_count(subject, predicate, object_)
    for current_page in range(1, pagination_data['num_pages'] + 1):
        params['page'] = current_page
        url = (
            requests.Request(
                url='https://query.wikidata.org/bigdata/ldf',
                params=params,
            )
            .prepare()
            .url
        )
        response = fetch_by_url(url, headers={'accept': 'application/ld+json'})
        json_data = response.json()

        for item in json_data['@graph']:
            if item['@id'].startswith('_:b'):
                continue
            if item['@id'].startswith('https://query.wikidata.org/bigdata/ldf'):
                continue
            if item['@id'].startswith('http://www.wikidata.org/.well-known/'):
                continue
            yield item

        # Bookkeeping
        del url, response, json_data


SCHEMA_ABOUT = urllib.parse.urlparse('http://schema.org/about')


def get_wikidata_concept_for_wikipedia_page(
    client,
    wikipage: str,
) -> wikidata.entity.Entity | None:
    triples = get_triples_internal(wikipage, SCHEMA_ABOUT, None)
    for item in list(triples):
        s = item['about'][3:]
        return client.get(s, load=False)
    return None


def get_triples(
    client,
    subject: Any = None,
    predicate: Any = None,
    object_: Any = None,
) -> Iterator[
    tuple[wikidata.entity.Entity, wikidata.entity.Entity, wikidata.entity.Entity]
]:
    iterator = get_triples_internal(subject, predicate, object_)
    for item in iterator:
        is_looking_for = item['@id'].startswith('wd:') and predicate.id in item
        if is_looking_for:
            s = subject
            if s is None:
                s = client.get(item['@id'][3:], load=False)
            o = object_ or item[predicate.id]
            yield (s, predicate, o)
        del item, is_looking_for


def get_backlinks(
    client,
    predicate: Any,
    object_: Any,
) -> Iterator[wikidata.entity.Entity]:
    for subject, _, _ in get_triples(
        client,
        subject=None,
        predicate=predicate,
        object_=object_,
    ):
        yield subject
